Classical Foreground Module
Donya Quick
Based on work from 690/691 (Master's thesis): 
http://haskell.cs.yale.edu/?post_type=publication&p=260

Last modified 24-Apr-2014


> module ClassicalFG where
> import GrammarMonad 
> import GrammarRules
> import GrammarRulesV
> import System.Random
> import Euterpea
> import ChordSpaces hiding (i)
> import PostProc
> import Data.List
> import FGTypes
> import Search


> data CConstants = CConstants {
>     ntLimC :: Int, -- limit for neighboring tone distance
>     ptLimC :: Int, -- limit for passing tone distance
>     pHalfC :: Double, -- probability of dividing a note's duration in half (alternative is x-en and en)
>     pTieC :: Double, -- probability of tying two identical notes
>     rootBassThreshC :: Double, -- probability of enforcing that the bass be the root
>     noCPLThreshC :: Int } -- voice-leading maximum, setting to 0 forces nearest neighbor fallback


> defConsts = CConstants 2 3 0.5 0.5 0.8 7

FG GENERATION

 

Get all pitches shared between the scales of two TNotes

> allPs :: TNote2 -> TNote2 -> [AbsPitch]
> allPs t1 t2 = 
>     let (o1, o2) = (ap t1 `div` 12, ap t2 `div` 12)
>         [oMin, oMax] = sort [o1, o2]
>         offs = map (12*) [oMin-1, oMin, oMax, oMax+1]
>         (s1, s2) = (baseScale $ tkey t1, baseScale $ tkey t2)
>     in  nub $ concatMap (\o -> t o [s | s<-s1, elem s s2]) offs where
>     baseScale :: Key -> [AbsPitch]
>     baseScale (k,m) = normOP $ t k (getScale m)


> type ForeFun = StdGen -> TNote2 -> TNote2 -> (StdGen, Maybe AbsPitch)

> pickPT :: AbsPitch -> ForeFun
> pickPT lim g t1 t2 = 
>     let [pMin, pMax] = sort [ap t1, ap t2]
>         f x = x>pMin && x<pMax && (x - pMin <=lim || pMax - x <=lim)
>         ps' = [x | x<-allPs t1 t2, f x]
>         (iNew, g') = randomR (0, length ps' - 1) g
>     in  if pMin == pMax || null ps' then (g, Nothing)
>         else (g', Just $ ps' !! iNew)

> pickNT :: AbsPitch -> ForeFun
> pickNT lim g t1 t2 = 
>     let [pMin, pMax] = sort [ap t1, ap t2]
>         f x =  (x < pMin && pMin - x <= lim) || (x > pMax && x - pMax <= lim)
>         ps' = [x | x<-allPs t1 t2, f x]
>         (iNew, g') = randomR (0, length ps' - 1) g
>     in  if pMin == pMax || null ps' then (g, Nothing)
>         else (g', Just $ ps' !! iNew)

> anticip, rept, doNothing :: ForeFun
> anticip g t1 t2 = (g, Just $ ap t2)
> rept g t1 t2 = (g, Just $ ap t1)
> doNothing g t1 t2 = (g, Nothing)

> f1 = pickPT . ptLimC 
> f2 = pickNT . ntLimC 
> [f3, f4, f5] = map const [anticip, rept, doNothing]

> allFFs :: CConstants -> [[(Double, ForeFun)]]
> allFFs c = 
>    [[(0.3, f1 c), (0.1, f2 c), (0.6, f5 c)], -- S
>    [(0.3, f1 c), (0.7, f5 c)], -- A
>    [(0.1, f1 c), (0.9, f5 c)], -- T
>    [(1.0, f5 c)]] -- B

> replaceIth :: [a] -> Int -> a -> [a]
> replaceIth xs i x = 
>     if i < 0 || i > length xs - 1 then error "Bad replacement index."
>     else take i xs ++ [x] ++ drop (i+1) xs

> splitP :: CConstants -> StdGen -> AbsPitch -> TNote2 -> (StdGen, [TNote2])
> splitP consts g newP t = 
>     let (r, g') = randomR (0,1.0::Double) g
>         dNew = if r < pHalfC consts then tdur t / 2 else en
>     in  (g', [t{tdur = tdur t - dNew}, t{tdur = dNew, ap = newP}])

Here we assume that the contexts include the voice in question. The |i| argument is the voice number.
The function only returns modifications of the first chord.

> addFgToVoice :: CConstants -> [(Double, ForeFun)] -> StdGen -> [TNote2] -> (StdGen, [TNote2])
> addFgToVoice c foreFuns g (t1:t2:ts) =
>     let (j, g1) = randomR (0, 1.0) g
>         fFun = chooseFF j foreFuns
>         (g2, t1') = applyForeFun c g1 t1 t2 fFun
>         (g3, tRest) = addFgToVoice c foreFuns g2 (t2:ts)
>     in  (g3, t1' ++ tRest) where
>     applyForeFun c g t1 t2 fFun = 
>         let (g1, newP) = fFun g t1 t2
>         in  case newP of Nothing -> (g1, [t1])
>                          Just x -> splitP c g1 x t1
> addFgToVoice c foreFuns g x = (g, x)

After adding foreground elements, ties can be considered. The following |stochTie| 
function stochastically ties notes in a voice. 

> stochTie :: CConstants -> StdGen -> [TNote2] -> (StdGen, [TNote2])
> stochTie consts g (t1:t2:ts) = 
>     let (r, g1) = randomR (0, 1.0::Double) g
>         (g2, (t2':ts')) = stochTie consts g1 (t2:ts)
>         (d1,d2') = (tdur t1, tdur t2')
>     in  if ap t1 == ap t2' && r < pTieC consts then (g2, t1 {tdur = d1 + d2'} : ts')
>         else (g2, t1 : t2' : ts')
> stochTie consts g ts = (g, ts)

Finally, the |addFG| function puts all of these elements together.

> addFG :: CConstants -> StdGen -> [[TNote2]] -> (StdGen, [[TNote2]])
> addFG c g vs = let (g', vs') = fgRec c g 0 vs in tieRec c g' vs' where
>     fgRec c g i vs = if i >= length vs then (g, vs) else 
>         let (g', v') = addFgToVoice c (allFFs c !! i) g (vs !! i)  
>             vs' = replaceIth vs i v'
>         in  fgRec c g' (i+1) vs'
>     tieRec c g [] = (g, [])
>     tieRec c g (v:vs) = 
>         let (g1,v') = stochTie c g v
>             (g2,vs') = tieRec c g1 vs
>         in  (g2, v':vs')

=============================================

INTERFACE TO TERM

> classicalFG :: StdGen -> Key -> Term CType -> (StdGen, (Music Pitch, Music Pitch))
> classicalFG g (k,m) t = 
>    let aChords = atTrans k $ toAbsChords (expand [] t) m
>        justChords = map (\(a,b,c) -> c) aChords
>        (g1,g2) = split g
>        (g3, eqs) = classBass 0.8 g2 $ map (eqClass satbOP opcEq) justChords
>        csChords = greedyProg' (noCPL 7) nearFall g3 eqs 
>        aChords' = zipWith (\(a,b,c) d -> (a,b,d)) aChords csChords
>        (g4,csFG) = addFG defConsts g2 $ reverse $ toTNote2 aChords'
>        is = [Bassoon, EnglishHorn, Clarinet, Oboe] 
>        fgM = vsToMusicI is $ reverse $ backToVoices csFG
>        csM = vsToMusicI is $ toVoices aChords'
>    in  (g4, (csM, fgM))

The following filter enforces that the bass holds the root with 
a certain probability (the "thresh" value). If the constraints can't 
be met, the bass is allowed to deviate from this rule.

> classBass :: Double -> StdGen -> [EqClass AbsChord] -> (StdGen, [EqClass AbsChord])
> classBass thresh g [] = (g, [])
> classBass thresh g (e:es) = 
>     let (r,g') = randomR (0,1.0::Double) g
>         e' = if r > thresh then e else filter rootFilter e
>         e'' = if null e' then e else e'
>         (g'', es') = classBass thresh g es
>     in  (g'', e'':es')

> rootFilter :: Predicate AbsChord
> rootFilter x = or $ map (opcEq x) [[0,0,4,7], [0,0,3,7], [0,0,3,6]]

The following version handles constraint satisfaction differently. Two MIDI
files are produced, one without melodic elements and one with them.

> classicalFG2 :: StdGen -> Key -> Term CType -> FilePath -> FilePath -> IO ()
> classicalFG2 g (k,m) t fn1 fn2 = do
>    let aChords = atTrans k $ toAbsChords (expand [] t) m
>        justChords = map (\(a,b,c) -> c) aChords
>        (g1,g2) = split g
>        qSpace = satbOP' g1
>        ecs = map (eqClass qSpace opcEq) justChords
>        cons = findInds [] t
>    (x, csChords) <- findSoln2 cons (progL 10) ecs -- DO CONSTRAINT SAT HERE
>    let aChords' = zipWith (\(a,b,c) d -> (a,b,d)) aChords csChords
>        (g4,csFG) = addFG defConsts g2 $ reverse $ toTNote2 aChords'
>        is = [Bassoon, EnglishHorn, Clarinet, Oboe] 
>        fgM = vsToMusicI is $ reverse $ backToVoices csFG
>        csM = vsToMusicI is $ toVoices aChords'
>    writeMidi fn1 fgM 
>    writeMidi fn2 csM



=============================================

UTILITY FUNCTIONS

> infInts g = let (i,g') = next g in i : infInts g'
> toVs2 = map (map f) where f ((k,m),d,p) = (p, d, m, k)
> toVs1 = map (map f) where f (p,d,m,k) = ((k,m),d,p)
> chooseFF j [x] = snd x
> chooseFF j ((p,x):t) = if j<p then x else chooseFF (j-p) t
> chooseFF j [] = error "(chooseFF) Nothing to choose from!"
