> {-# OPTIONS_GHC -XBangPatterns #-}

> module InOut2 where
> import Data.List
> import CykParser
> import Parser hiding (testStr, testStr2)
> import System.Random
> import Control.Parallel.Strategies

Donya Quick
Last modified: 20-Dec-2013

CHANGES:
- Modified to handle template-style rules
- Modified to handle multiple strings
- Modified to handle A -> BCD
- Modified to handle A -> B synonyms - NEED TO CHECK THIS AT START SYM LEVEL
- Modified to allow for minimum production probs
- Removed redundant parse tables from type signatures of Env/EnvM
- Added support for A -> A identity rules.


> prod :: (Num a, Eq a) => [a] -> a
> prod [] = 1
> prod (x:xs) = if x==0 then 0 else x * prod xs

CONSTANTS

> minProbT = 0.00001 -- set to >0 to avoid zero probability rules 

> type TSym a = (a, Rational)
> type RHS a = [TSym a]
> type TRule a = (a, RHS a)
> type TStr a = [TSym a]

> data TEnv a = TEnv {rules :: [TRule a], 
>                     phis :: [Double],
>                     tAlph :: [Rational],
>                     startSym :: TSym a} 

> type Alpha a = ((a, Int, Int), Double)
> type Beta a = ((a, Int, Int), Double)


> phi :: (Show a, Eq a) => TEnv a -> TRule a -> Double
> phi env r = case findIndex (== (fst r, normR $ snd r)) (rules env) of
>     Just i -> phis env !! i
>     Nothing -> error ("No such rule! "++ show r)

> idRule :: (Eq a) => TRule a -> Bool
> idRule (lhs, rhs) = [lhs] == map fst rhs

> normR :: RHS a -> RHS a
> normR xs = 
>     let denom = sum $ map snd xs
>     in  map (\(a,b) -> (a, b / denom)) xs

> scaleR :: Rational -> RHS a -> RHS a
> scaleR d = map (\(a,b) -> (a, d*b))

=======================

ALPHAS - SEEMS OK

> alpha :: (Eq a, Show a) => [Alpha (TSym a)] -> TEnv a -> TStr a -> TSym a -> Int -> Int -> Double
> alpha as env xs (a,t) i j = 
>     let inTerms b c k = prod [phi env (a, normR [b,c]), alpha as env xs b i k, 
>                         alpha as env xs c (k+1) j]
>         inSum [b] = prod [phi env (a, normR [b]), alpha as env xs b i j]
>         inSum [b,c] = sum $ map (inTerms b c) [i..j-1]
>         inSum [b,c,d] = sum $ map (\(k,l) -> 
>                         prod [phi env (a, normR [b,c,d]), alpha as env xs b i k, 
>                         alpha as env xs c (k+1) l, alpha as env xs d (l+1) j])
>                         [(k,l) | k<-[i..j], l<-[i..j], k<l, l<j]
>         vals = filter ((/=[a]).map fst) $ genDown env (a,t)
>         retVal = if i==j && (xs!!i) == (a,t) then 1.0 
>                  else sum $ map inSum vals
>     in  case lookup ((a,t),i,j) as of 
>             Nothing -> retVal
>             Just p -> p

vals is: [rhs | (lhs,rhs) <- rules, lhs==a, [lhs] /= rhs]

> genDown :: (Eq a) => TEnv a -> TSym a -> [RHS a]
> genDown env (a,t) = 
>     let rhs0 = [rhs | (lhs, rhs) <- rules env, lhs == a]
>     in  map (scaleR t) rhs0

> distComp (x,a,b) (y,c,d) = compare (b-a) (d-c)

> makeAlphas :: (Eq a, Show a) => TEnv a -> TStr a -> [Alpha (TSym a)]
> makeAlphas env xstr = 
>     let n = length xstr
>         ijs = [(i,j) | i<-[0..n-1], j<-[0..n-1], i<=j]
>         aVals = [((a,t),i,j) | a<-nub (map fst $ rules env), t<-(tAlph env), (i,j)<-ijs]
>         aijs = reverse $ sortBy distComp aVals
>     in  filter ((>0).snd) $ recAlpha env [] aijs where
>     recAlpha env alphas [] = alphas
>     recAlpha env alphas (ax@(a,i,j):axs) = 
>         let aNew = alpha alphas env xstr a i j
>         in  recAlpha env ((ax,aNew):alphas) axs

> makeAlphasM :: (Eq a, Show a) => TEnv a -> [TStr a] -> [[Alpha (TSym a)]]
> makeAlphasM env = map (makeAlphas env)

> alpha' :: (Eq a) => [Alpha (TSym a)] -> TSym a -> Int -> Int -> Double
> alpha' as a i j = case lookup (a,i,j) as of -- TO BE USED DURING FURTHER CALCS
>     Nothing -> 0.0
>     Just p -> p


=========

BETAS - HAS PROBLEMS!

> beta2 :: (Eq a, Show a) => [Alpha (TSym a)] -> [Beta (TSym a)] -> TEnv a -> 
>     TStr a -> TSym a -> Int -> Int -> Double
> beta2 as bs env xs a i j = 
>     let n = length xs
>         inTerms1 b c k = prod [phi env (b, normR [c,a]), alpha as env xs c k (i-1), 
>                          beta as bs env xs (b, sum $ map snd [c,a]) k j]
>         inTerms2 b c k = prod [phi env (b, normR [a,c]), alpha as env xs c (j+1) k, 
>                          beta as bs env xs (b, sum $ map snd [a,c]) i k]
>         synCase (b, [a']) = prod [phi env (b, normR [a]), beta as bs env xs (b, snd a) i j] 
>         inSum1 (b,[c,a']) = sum $ map (inTerms1 b c) [0..i-1] -- CHECK INDEXING
>         inSum2 (b,[a',c]) = sum $ map (inTerms2 b c) [j+1..n-1] -- CHECK INDEXING
>  --  USE OF SYNONYM FUNCTION BELOW MAY OR MAY NOT BE CORRECT --
>     in  if i==0 && j == n - 1 && a==startSym env then 1.0 
>         else sum (map inSum1 $ genUp env a 1 2) + -- (b, [c,a])
>              sum (map inSum2 $ genUp env a 0 2) + -- (b, [a,c])
>              sum (map synCase $ genUp env a 0 1)  -- (b, [a])


> genUp :: (Eq a) => TEnv a -> TSym a -> Int -> Int -> [TRule a] 
> genUp env (a,t) i n = 
>     let rCands = [r | r<-rules env, length (snd r) == n, (map fst $ snd r) !! i == a]
>         rScaled = map (fixTs t i) rCands 
>         rOK = filter (\(l,r) -> and $ map ((`elem` tAlph env).snd) r) rScaled
>     in  filter (`elem` rules env) $ filter (not . idRule) rOK -- map (\r -> (i,n,r)) rOK

> fixTs t i (lhs, rhs) = 
>     let factor = t / (map snd rhs !! i)
>     in  (lhs, map (\(x,y) -> (x, factor*y)) rhs)


BETA FUNCTION THROWS BAD RULE EXCEPTIONS
GIVING BAD DURATIONS AND WEIRD COMBOS
Ex: (I, [(I, 1/3), (V, 1/3), (V, 1/3)])
Is fine with degree 2 rules, but not degree 3

> beta :: (Eq a, Show a) => [Alpha (TSym a)] -> [Beta (TSym a)] -> TEnv a -> 
>     TStr a -> TSym a -> Int -> Int -> Double
> beta as bs env xs a i j = 
>     let n = length xs 
>         inTerms1 b c d (l,k) = prod [phi env (b, normR [c,d,a]), alpha' as c l (k-1), -- alpha as env xs c l (k-1), 
>                                alpha' as d k (i-1), -- alpha as env xs d k (i-1), 
>                                beta as bs env xs (b, sum $ map snd [c,d,a]) l j]
>         inTerms2 b c d (l,k) = prod [phi env (b, [c,a,d]), alpha' as d (j+1) k, -- alpha as env xs d (j+1) k,
>                                alpha' as c l (i-1), -- alpha as env xs c l (i-1), 
>                                beta as bs env xs (b, sum $ map snd [c,a,d]) l k]
>         inTerms3 b c d (l,k) = prod [phi env (b, [a,c,d]), alpha' as d (l+1) k, -- alpha as env xs d (l+1) k,
>                                alpha' as c (j+1) l, -- alpha as env xs c (j+1) l, 
>                                beta as bs env xs (b, sum $ map snd [a,c,d]) i k]
>         lks1 = [(l,k) | l<-[0..n-1], k<-[0..n-1], l >= 0 && l<k && k<i]
>         lks2 = [(l,k) | l<-[0..n-1], k<-[0..n-1], l >= 0 && l<i && k>j]
>         lks3 = [(l,k) | l<-[0..n-1], k<-[0..n-1], l>j, k>l]
>         inSum1 (b,[c,d,a']) = sum $ map (inTerms1 b c d) lks1
>         inSum2 (b,[c,a',d]) = sum $ map (inTerms2 b c d) lks2
>         inSum3 (b,[a',c,d]) = sum $ map (inTerms3 b c d) lks3
>         retVal = 
>              sum (map inSum1 $ genUp env a 2 3) +
>              sum (map inSum2 $ genUp env a 1 3) +
>              sum (map inSum3 $ genUp env a 0 3) +
>              beta2 as bs env xs a i j -- HANDLE LENGTH 2 AND 1
>    in   case lookup (a,i,j) bs of Nothing -> retVal
>                                   Just p -> p 


> makeBetas :: (Eq a, Show a) => [Alpha (TSym a)] -> TEnv a -> TStr a -> [Beta (TSym a)]
> makeBetas as env xstr = 
>     let n = length xstr
>         ijs = [(i,j) | i<-[0..n-1], j<-[0..n-1], i<=j]
>         aVals = [((a,t),i,j) | a<-nub (map fst $ rules env), t<-(tAlph env), (i,j)<-ijs]
>         aijs = sortBy distComp aVals
>     in  filter ((>0).snd) $ recBeta as env [] aijs where
>     recBeta as env betas [] = betas
>     recBeta as env betas (x@(a,i,j):xs) = 
>         let bNew = beta as betas env xstr a i j
>         in  recBeta as env ((x,bNew):betas) xs

> beta' :: (Eq a) => [Beta (TSym a)] -> TSym a -> Int -> Int -> Double
> beta' bs a i j = case lookup (a,i,j) bs of -- TO BE USED DURING FURTHER CALCS
>     Nothing -> 0.0
>     Just p -> p

> makeBetasM :: (Eq a, Show a) => [[Alpha (TSym a)]] -> TEnv a -> [TStr a] -> [[Beta (TSym a)]]
> makeBetasM as env = zipWith (\a x -> makeBetas a env x) as


=======================

MU STUFF

> mu2 :: (Show a, Eq a) => [Alpha (TSym a)] -> [Beta (TSym a)] -> TEnv a -> TStr a ->
>     TRule a -> Int -> Int -> Int -> Double
> mu2 as bs env xs r@(a,[b]) i k j = 
>     if i==k || k == j then -- spanning 1+ symbols
>         let [b'] = fixT2 xs i j [b] in -- correct for duration span 
>              prod [beta' bs (a, snd b') i j,      
>                    phi env (a, normR [b]),                
>                    alpha' as (fst b, sum $ map snd (take (j-i+1) xs)) i j] 
>     else error ("(mu) Bad rule: "++show (r,i,k,j))
> mu2 as bs env xs r@(a,[b,c]) i k j = 
>   let [b',c'] = fixT2 xs i j [b,c] in -- correct for duration span 
>     prod [beta' bs (a, snd b' + snd c') i j, 
>           phi env (a, normR [b,c]),
>           alpha' as b' i k, 
>           alpha' as c' (k+1) j]

> fixT2 xs i j = 
>     let factor = sum (map snd (take (j-i+1) $ drop i xs))
>     in  map (\(x,y) -> (x, y*factor))

map ((\(x,y) -> (x, y* sum (map snd [b,c])))) [b,c]

> mu :: (Show a, Eq a) => [Alpha (TSym a)] -> [Beta (TSym a)] -> TEnv a -> TStr a ->
>     TRule a -> Int -> Int -> Int -> Int -> Double
> mu as bs env xs r i k l j = 
>     case r of 
>         (a,[b,c,d]) -> 
>            let [b',c', d'] = fixT2 xs i j [b,c,d] -- correct for duration span
>            in  prod [phi env (a, normR [b',c',d']), 
>                      beta' bs (a, sum $ map snd [b',c',d']) i j, 
>                      alpha' as b' i k,
>                      alpha' as c' (k+1) l, 
>                      alpha' as d' (l+1) j]
>         _ -> mu2 as bs env xs r i k j 


======================

COUNTS

> count as bs env xs r@(a,[b]) = let n = length xs in
>     sum (map (\(i,j) -> mu2 as bs env xs r i i j ) 
>     [(i,j) | i<-[0,n-1], j<-[0..n-1], i<=j]) / z as bs env xs
> count as bs env xs r@(a,[b,c]) = 
>     let n = length xs
>         term (i, k, j) = mu2 as bs env xs r i k j / z as bs env xs
>         triples = [(i,k,j) | i<-[0..n-2], k<-[0..n-1], j<-[1..n-1], i <= k && k < j]
>     in  sum $ map term triples
> count as bs env xs r@(a,[b,c,d]) = --ADDED CASE
>     let n = length xs
>         term (i, k, l, j) = mu as bs env xs r i k l j / z as bs env xs
>         quads = [(i,k,l,j) | i<-[0..n-2], k<-[0..n-1], l<-[0..n-1], 
>                    j<-[1..n-1], i <= k && k<l && l < j]
>     in  sum $ map term quads
> count as bs env xs r = error ("(count) Badly formed rule: "++show r)

> z :: (Eq a) => [Alpha (TSym a)] -> [Beta (TSym a)] -> TEnv a -> TStr a -> Double
> z as bs env xs = alpha' as (startSym env) 0 (length xs-1)

> fLast :: (Show a, Eq a) => [[Alpha (TSym a)]] -> [[Beta (TSym a)]] -> TEnv a -> [TStr a] ->
>     TRule a -> Double
> fLast as bs env xs r = 
>     let exprs = map (\(a,b,x) -> count a b env x r) $ zip3 as bs xs
>     in  sum $ seq (as, bs, xs) (exprs `using` parList rdeepseq) 


===========================

RE-ESTIMATION STUFF

> newQT' :: (Eq a, Show a) => [[Alpha (TSym a)]] -> [[Beta (TSym a)]] -> TEnv a -> [TStr a] ->
>     TRule a -> Double
> newQT' as bs env xs r@(lhs,rhs) = 
>     let f rx = fLast as bs env xs rx 
>         rs = filter (\(lhs',rhs') -> lhs == lhs') $ rules env -- CHECK CORRECTNESS
>         denom = sum (map f rs)
>     in  if denom == 0 then f r else f r / denom


====================================

> normalizeT :: (Eq a, Show a, Show b) => Double -> [(a,b)] -> [Double] -> [Double]
> normalizeT minP rules vals =
>     let newVals = map (\x -> if x < minP || x <= 0 then minP else x) vals 
>     in  map (norm2 (zip rules newVals)) (zip rules newVals)

> norm2 :: (Eq a, Show a, Show b) => [((a,b), Double)] -> ((a,b), Double) -> Double
> norm2 rules ((l,r), d) = 
>     let denom = sum $ map snd $ filter (\((l',r'), d') -> l==l') rules
>     in  if denom > 0 then d / denom else error ("(norm2) Bad rule: "++ show (l,r, d)) 

> infRand :: StdGen -> [Double]
> infRand g = 
>     let (d,g') = randomR (0.0, 1.0) g
>     in  d : infRand g'

> randQFun2 g rules (l,r) =
>     let ds = take (length rules) (infRand g) -- get a bunch of random values
>         rds = zip rules ds
>         d = case findIndex (==(l,r)) rules of Just i -> snd (rds !! i)
>                                               Nothing -> 0.0 -- error "(randQFun) Rule not found!"
>         rs = filter (\((l',r'),d') -> l==l') rds
>         newDenom = sum (map snd rs)
>     in  d / newDenom
	 
> randQFunT g rules = map (randQFun2 g rules) rules
> uniQsT rs = normalizeT 0.0 rs (let n = length rs in take n $ repeat (1))


====================================

> learnProbsT :: (Eq a, Show a) => TEnv a -> [TStr a] -> Int -> Double -> Double -> IO [[Double]]
> learnProbsT e xs i minP dThresh = if i<=0 then return [] else do
>     let as = makeAlphasM e xs
>     --putStrLn "Alphas..."
>     --putStrLn (show (head $ head as) ++ " ... " ++ show (last $ last as))
>     let bs = makeBetasM as e xs
>     --putStrLn "Betas..."
>     --putStrLn (show (head bs) ++ " ... " ++ show (last bs))
>     putStrLn "Calculating production probs..."
>     let qs0 = phis e --qf -- map qf r 
>         qs1s = map (\rx -> newQT' as bs e xs rx) $ rules e
>         qs1 = seq (as, bs, e) (qs1s `using` parList rdeepseq) -- parallelize computation of qs1
>         qf' = if minP > 0.0 then normalizeT minP (rules e) qs1 else qs1
>         e' = e{phis=qf'}-- (r,s,strs,qf') 
>         diff = totalDiff qs0 qf'
>     putStrLn ("Total probability mass changed: "++show diff)
>     sequence $ map (\q -> putStr (show q ++ ", ")) qf' 
>     let badInds = findIndices (\x -> x<=0) qs1
>         badStr = if null badInds then "ok" 
>                  else "Zero-prob rules: "++show badInds
>         (iNext, skipping) = if diff <= dThresh then (0, True) else (i-1, False)
>         itrStr = if skipping then ("Q has converged with "++show (i-1)++" iterations left.")
>                  else ("Iterations left: "++show (i-1))
>     putStrLn itrStr
>     qNext <- learnProbsT e' xs iNext minP dThresh 
>     return (qf' : qNext)

> totalDiff :: [Double] -> [Double] -> Double
> totalDiff d1 d2 = sum $ map abs $ zipWith subtract d1 d2 

====================================

TESTING

data TEnv a = TEnv {rules :: [TRule a], 
                    phis :: [Double],
                    tAlph :: [Rational],
                    startSym :: TSym a} 

asdf

> rulesx = [(1, [(2, 0.5), (1, 0.5)]),
>          (1, [(1, 1.0)]),
>          (1, [(2, 1.0)]), -- PROBLEM WITH THIS RULE
>          (2, [(2, 0.5), (1, 0.5)])]

> tsx = [1.0, 0.5, 0.25] :: [Rational]

> qs0x = [0.5, 0.5, 0.5, 0.5] :: [Double]

> strx = [(2, 0.5), (2, 0.25), (1, 0.25)] :: TStr Int

> envx = TEnv rulesx qs0x tsx (1, 1.0) :: TEnv Int
 
> asx = makeAlphas envx strx
> bsx = makeBetas asx envx strx
> qsx = map (newQT' [asx] [bsx] envx [strx]) rulesx

> envx' = envx {phis = qsx}
> asx' = makeAlphas envx' strx
> bsx' = makeBetas asx' envx' strx
> qsx' = map (newQT' [asx'] [bsx'] envx' [strx]) rulesx


> myTest = learnProbsT envx [strx] 10 0.0001 0.01 





