> module Markov where
> import System.Random
> import GrammarMonad
> import Euterpea
> import PostProc
> import Data.List
> import ChordSpaces
> import Control.Parallel.Strategies

> randStep :: [a] -> (StdGen, [a]) -> (StdGen, [a])
> randStep states (g, xs) = 
>     let (i,g') = randomR (0, length states - 1) g 
>     in  (g', states !! i : xs)  

> randSeq :: [a] -> (StdGen, [a]) -> [(StdGen, [a])]
> randSeq states = iterate (randStep states)

> randData :: (Int,Int) -> [a] -> [b] -> StdGen ->[[(a,b)]]
> randData range states1 states2 g = 
>     let (len, g1) = randomR range g
>         (g2, xs) = randSeq states1 (g1, []) !! len
>         (g3, ys) = randSeq states2 (g2, []) !! len
>     in  zip xs ys : randData range states1 states2 g3



================================

> allTriads = concatMap f [[0,4,7], [0,3,7], [0,3,6]] where
>     f c = map (\i -> map (+i) c) [0..11]

> isChord x = elem (normOPC x) $ map normOPC allTriads

> cFilter x = isChord x && x == sort x

> mkQSpaces cs g = 
>     let (g1, g2) = split g
>         q = randomize g1 cs // opEq
>     in  q : mkQSpaces cs g2

======================

ADDING FOREGROUND STUFF

> procNote :: StdGen -> TNote -> (StdGen, [TNote])
> procNote g (k,d,a) =
>     let (p, g1) = randomR (0.0, 1.0 :: Double) g
>         (i, g2) = randomR (-5, 5) g
>         newNotes = [(k,d/2,a), (k,d/2,a+i)]
>     in  if p <= pSplit then (g2, newNotes) else (g2, [(k,d,a)])

> addFG :: StdGen -> [TChord] -> (StdGen, [Voice])
> addFG g ts = let vs = toVoices ts in recVoice g vs where
>     recVoice :: StdGen -> [Voice] -> (StdGen, [Voice])
>     recVoice g [] = (g, [])
>     recVoice g (v:vs) = 
>         let (g1, vs') = recVoice g vs
>             (g2, v') = fgVoice g1 v
>         in  (g2, v':vs')
>     fgVoice :: StdGen -> Voice -> (StdGen, Voice)
>     fgVoice g [] = (g, [])
>     fgVoice g (t:ts) = 
>         let (g1, ts') = fgVoice g ts
>             (g2, t') = procNote g1 t
>         in  (g2, t' ++ ts')


================================

Concrete data values


> range = (8,16) -- NEED TO CHECK ON THIS
> qRanges = lgRanges -- NEED TO CHECK ON THIS
> seed1 = 0 -- for generating Roman numerals
> seed2 = 1 -- for generating randomized quotient spaces
> seed3 = 2 -- for generating chord space walks
> seed4 = 3 -- for generating foreground walks
> pSplit = 0.15 -- for adding foreground notes

> lgRanges = [(40,60), (47,67), (52,76), (60,81)] 
> smRanges = map (\(l,u) -> (l+12, u+12)) [(40,50), (48,58), (56,64), (62,72)]

> type MTerm a = ((AbsPitch, Mode), a)

> allCs = [I, II, III, IV, V, VI, VII]
> allKMs = [(k,m) | k<-[0..11], m<-[Major, Minor]]

> allStates :: [MTerm CType]
> allStates = [(km, c) | km <- allKMs, c<-allCs]

> allDurs = [qn,qn,qn,hn]

> randChords :: Int -> (Int, Int) -> [[(MTerm CType, Dur)]]
> randChords seed range = randData range allStates allDurs $ mkStdGen seed

> theCChords = take 100 $ filter ((==4) .sum . map snd) $ randChords seed1 range
> convert ((km,x),d) = (km, d, x)
> theAChords1 = map (map (cChordToA.convert)) theCChords

> (theDurs1, theAs1) = (map (map fst) xs, map (map snd) xs) where
>     xs = map (map (\(k,d,a) -> ((k,d),a))) theAChords1


> allChords = filter cFilter $ makeRange qRanges
> qSpaces = mkQSpaces allChords $ mkStdGen seed2

greedyProg qs r c f g (x:xs)

> theAs2 = map f $ zip theAs1 qSpaces where
>     f (cs, qSpace) = greedyProg qSpace opcEq (noCPL 7) nearFall (mkStdGen seed3) cs 

> theAChords2 = zipWith (zipWith (\(k,d) a -> (k,d,a))) theDurs1 theAs2 :: [[TChord]]

> theVoices = map (addFG $ mkStdGen seed4) theAChords2

> instrs = [Bassoon, EnglishHorn, Clarinet, Oboe]

> theMusics = map (vsToMusicI instrs . snd) theVoices

> allMusic = line $ intersperse (rest 2) theMusics

> fileBase = "random\\rand"

> theOutputs = zipWith f theMusics [0..] where
>     f m i = putStrLn ("Writing example" ++ show i ++ "...") >> 
>             writeMidi (fileBase ++ show i ++ ".mid") m 



