> module TemporalGen where
> import GrammarMonad
> import GrammarRules -- 2
> import Euterpea
> import Data.List
> import System.Random
> import InOut3
> import RuleUtils hiding (minDur)


> type TSym a = (a, Rational)
> type RHS a = [TSym a]
> type TRule a = (a, RHS a)


> tSeed = i maxDur :: Term CType
> maxDur = 1.0
> minDur = sn
> dataPoints = 1000
> rSeed = 0
> iters = 4
> useLets = False 

> infInts g = 
>     let (g1, g2) = split g
>         (i, g') = next g
>     in  i : infInts g2

> genData = 
>     let f i = gen (ruleSet minDur useLets) iters i 0 tSeed
>         theData = map f $ take dataPoints $ infInts (mkStdGen rSeed)
>     in  map toPairs theData

> writeData = writeFile "tdata.txt" $ concat $ intersperse "\n" $ map show genData


======================

> convRule :: (Prob, TRule CType) -> Rule CType
> convRule (p, (lhs, rhs)) = (lhs, p) :-> toRuleFun rhs

> toRuleFun :: [(CType, Dur)] -> RuleFun CType
> toRuleFun rhs ctxt t = S $ map (\(c,d) -> (NT $ Chord (d*t) c)) rhs

> readTemporalRules :: FilePath -> IO ((CType, Dur), [(Prob, TRule CType)])
> readTemporalRules fp = do
>     str <- readFile fp
>     let theLines = lines str
>         startSym = read (head theLines) :: (CType, Dur)
>         theRules = map (read) $ tail $ theLines
>     return (startSym, theRules)

> readTemporalRules2 :: FilePath -> IO ((CType, Dur), [Rule CType])
> readTemporalRules2 fp = do
>     (startSym, theRules) <- readTemporalRules fp
>     return (startSym, map convRule theRules)

> readPTGG = readTemporalRules2


> readTemporalData :: FilePath -> IO [[(CType, Dur)]]
> readTemporalData fp = do
>     str <- readFile fp
>     return $ map (read) $ lines str

> showTemporal :: ((CType, Dur), [TRule CType]) -> String
> showTemporal (startSym, rules) = 
>     show startSym ++ "\n\n" ++
>     concat(intersperse "\n" $ map showTRule rules)

> showTRule (lhs, rhs) = show lhs ++ "  ->  "++ 
>     concat (intersperse "  " $ map show rhs)

> printTemporal = putStrLn . showTemporal
 
> genFromFile fpIn fpOut minD dataAmt seed numIters = do
>     (startSym, rs) <- readTemporalRules2 fpIn
>     let  f i = gen (relRuleSet minD rs) numIters i 0 tSeed
>          theData = map f $ take dataPoints $ infInts (mkStdGen seed)
>          theData' = map toPairs theData
>     writeFile fpOut $ concat $ intersperse "\n" $ map show theData'

> genDataTest = genFromFile "trules.txt" "tdata.txt" 0.01 1000 0 3

> genDataTest2 = genFromFile "trules2.txt" "tdata2.txt" 0.01 1000 0 3


> trules2 = genFromFile "trules2.txt" "tdata2.txt" minDur dataPoints 4 iters

> trules3 = genFromFile "learning\\trules3.txt" "learning\\tdata3.txt" minDur dataPoints 4 iters
