> module Foreground where
> import Euterpea
> import System.Random
> import Data.List

> type NoteTerm = (AbsPitch, Dur, Mode, Key)
> type Voice2 = [NoteTerm]
> --data Mode = Major | Minor
> --	deriving (Show, Eq, Ord)
> type Progression = [[AbsPitch]]
> type Voicing = [Voice2]

> type Key = AbsPitch

> dtrans = Data.List.transpose

> toNTs :: Key -> Dur -> [[AbsPitch]] -> [Voice2]
> toNTs k d cs = 
>     let k' = k `mod` 12 
>         m = if k < 12 then Major else Minor
>         f a = (a, d, m, k')
>         vs = dtrans cs
>     in  map (map f) vs

> data NoteRule = NoteRule([Int], [Int], [Rational])
>	deriving (Show, Eq, Ord)

> type NoteRuleList = [(Mode, [NoteRule])]



==================

Code below taken from Constants.lhs in 690.

> restVal = -1001 :: AbsPitch
> restMax = -1000 :: AbsPitch


Code below taken from Utility.lhs in 690.

> contained :: Eq a => a -> [a] -> Bool
> contained a [] = False
> contained a ((a'):t) = if (a==a') then True else contained a t


> subSet :: Eq a => [a] -> [a] -> Bool
> subSet [] a = True
> subSet a [] = False
> subSet (a:as) bs = if contained a bs then subSet as bs else False

Search a list of (key, data) pairs.

> search :: (a -> a -> Bool) -> [(a,b)] -> a -> b
> search feq [] key = error "(search) Supplied key was not contained in the list."
> search feq (h:t) key = if feq key (fst h) then snd h else search feq t key 

> addToEnd :: a -> [a] -> [a]
> addToEnd a b = reverse (a : (reverse b))

Code below taken directly from Foreground2.lhs in 690.

============ CONTEXT FINDING =============

"Context" here refers to the notes (as scale indices) that are playing
concurrently during some period of time in a compostion. This is needed
to make good choices for adding/modifying notes.

Find the total duration of a Voice with the functino below.

> sumDur :: Voice2 -> Dur
> sumDur [] = 0
> sumDur ((i,d,st,k):ts) = d + sumDur ts


Get the scale indices appearing in a et of voices between particular
points in time (durations).

> getConIndices :: Dur -> Dur -> Dur -> [NoteTerm] -> [Int] -> [Int]
> getConIndices mind maxd currd [] is = is
> getConIndices mind maxd currd ((i,d,st,k):ts) is = 
>     if currd < mind then getConIndices mind maxd (currd + d) ts is else
>       if currd > maxd then is else 
>         getConIndices mind maxd (currd+d) ts (i:is)



Type definition for local use within this file:

> type VContext = [Int] -- list of notes playing at the same time

This function converts from indices considering octaves to those ignoring
octave (forced into the [0,11] range). This is a non-reversible process.


> scaleContext :: VContext -> Key -> VContext
> scaleContext con k = map (`mod` 12) (map (subtract k) con)



=============== NOTE RULE SELECTION ====================

Locate all matching note rules for a given context.

First some functions are neeeded to simplify data access:

> getNRContext :: NoteRule -> VContext
> getNRContext (NoteRule (c, n, p)) = c

> getNRNotes :: NoteRule -> VContext
> getNRNotes (NoteRule (c, n, p)) = n


Now we need a way to look up rules given a context:

> getMatchingRules :: VContext -> [NoteRule] -> [NoteRule]
> getMatchingRules [] n = n
> getMatchingRules cs [] = []
> getMatchingRules cs (n:ns) = 
>     if subSet (getNRContext n) cs then n : (getMatchingRules cs ns) else getMatchingRules cs ns 


The above process will give a lot of note restrictions. These 
need to be merged by intersecting the sets of allowable notes.

> mergeNoteOptions :: [NoteRule] -> [Int]
> mergeNoteOptions [] = []
> mergeNoteOptions [n] = getNRNotes n
> mergeNoteOptions (n:ns) = intersect (getNRNotes n) (mergeNoteOptions ns)


This function wraps the previous two:

> getAvailableIndices :: VContext -> [NoteRule] -> [Int]
> getAvailableIndices c ns = mergeNoteOptions (getMatchingRules c ns)


This function allows the selection of the nearest octvate-considering
index given a context and a reference note. The distance measure here
is just in scale degrees - it does not consider chord geometry currently.

> chooseNearestIndexB :: [Int] -> Int -> Int -> Int
> chooseNearestIndexB [] goal best = best
> chooseNearestIndexB (i:is) goal best = 
>     if abs (i - goal) < abs (best - goal) then chooseNearestIndexB is goal i else chooseNearestIndexB is goal best

> chooseNearestIndex :: [Int] -> Int -> Int
> chooseNearestIndex is goal = chooseNearestIndexB is goal 1000




This is a utility function to generate three octaves of a scale as
an index list.

> makeIndexSets :: [Int] -> [Int]
> makeIndexSets is = (map (\x -> x - 12) is) ++ is ++ (map (+12) is)


=================== FUNCTIONS ON NOTETERMS =========================

The functions in this section modify lists of NoteTerms to add foregrounds.

Change the duration of a term to some other value.

> changeDur :: NoteTerm -> Dur -> NoteTerm
> changeDur (i,d1,st,k) d2 = (i,d2,st,k)

Extract the duration of a NoteTerm. 

> getDur :: NoteTerm -> Dur
> getDur (i,d1,st,k) = d1


Get the scale index of a NoteTerm.

> getIndex :: NoteTerm -> Int
> getIndex (i,d,st,k) = i


Check whether a NoteTerm is a note.

> isNote :: NoteTerm -> Bool
> isNote (i,d,st,k) = i > restMax


> getMode :: NoteTerm -> Mode
> getMode (i,d,s,k) = s

> getKey :: NoteTerm -> Key
> getKey (i,d,s,k) = k


This is an updated version of the passing tone function that considers the context

> addPTone :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> addPTone (r:r1:rs) t1 t2 ctxt rulesx = 
>     let i1 = getIndex t1
>         i2 = getIndex t2
>         m = min (i1-k1) (i2-k1) -- normalize to first key
>         o = m `div` 12
>         rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail = (intersect iAvail1 iAvail2) 
>         range = i2 - i1
>         r' = if (range == 0) then 0 else r `mod` range
>         i3 = chooseNearestIndex iAvail ((m `mod` 12) + r')
>         i3' = (o * 12) + i3 + k1
>         d1' = getDur t1 / 2
>         d2' = getDur t2 / 2
>         r1' = r1 `mod` 2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>
> -- at this point, it is necessary to see if it is even possible to add a passing tone
>         ts = if i3' >= (max i1 i2) || i3' <= (min i1 i2) then [t1, t2] else 
>                if r1' == 0 then [changeDur t1 d1', (i3', d1',s1,k1), t2] else 
>                    [t1, (i3', d2',s2,k2), changeDur t2 d2']
>     in  (rs, ts)



This is an updated function that considers context:

> addNToneUp :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> addNToneUp (r:r1:rs) t1 t2 ctxt rulesx = 
>     let i1 = getIndex t1
>         i2 = getIndex t2
>         m = max (i1-k1) (i2-k1)
>         o = m `div` 12
>         r' = r `mod` 12
>         rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail = union ctxt1 (intersect iAvail1 iAvail2)
>         i3x = (o * 12) + (chooseNearestIndex iAvail (m `mod` 12)) -- r') 
>         i3y = chooseNearestIndex [i3x-12, i3x, i3x+12] m
>         i3 = i3y + k1
>         d1' = getDur t1 / 2
>         d2' = getDur t2 / 2
>         r1' = r1 `mod` 2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>         ts = if r1' == 0 then [changeDur t1 d1', (i3, d1',s1,k1), t2] else [t1, (i3, d2',s2,k2), changeDur t2 d2']
>     in  if (length iAvail > 0) then (rs, ts) else (rs, [t1,t2])






This is an updated version to consider context:

> addNToneDwn :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm]) 
> addNToneDwn (r:r1:rs) t1 t2 ctxt rulesx  = 
>     let i1 = getIndex t1
>         i2 = getIndex t2
>         m = min (i1-k1) (i2-k1)
>         o = m `div` 12
>         r' = r `mod` 12
>         rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail = union ctxt1 (intersect iAvail1 iAvail2)
>         i3x = (o * 12) + (chooseNearestIndex iAvail (m `mod` 12))
>         i3y = i3x -- if i3x < m then i3x + 12 else i3x
>         i3 = i3y + k1 
>         d1' = getDur t1 / 2
>         d2' = getDur t2 / 2
>         r1' = r1 `mod` 2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>         ts = if r1' == 0 then [changeDur t1 d1', (i3, d1',s1,k1), t2] else [t1, (i3, d2',s2,k2), changeDur t2 d2']
>         -- ts = [t1, (i3, d2',s2,k2), changeDur t2 d2']
>     in  if (length iAvail > 0) then (rs, ts) else (rs, [t1,t2]) -- (rs, ts)






This function will add an anticipation of the SECOND note (no equivalent for the first note).

NOTE: This is modified from the version in Foreground.lhs. Context-sensitivity is utilized
this time to avoid strange effects due to operating in pitch space vs. scale indices.

> addAnticip :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> addAnticip rands t1 t2 ctxt rulesx = 
>     let i3 = getIndex t2
>         d1' = getDur t1 / 2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>         ts = [changeDur t1 d1', (i3, d1',s1,k1), t2]
>         rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail' = (intersect iAvail1 iAvail2)
>         iAvail = if (length iAvail' > 0) then iAvail' else ctxt1
>         inds = iAvail 
>         isOk = contained (getIndex t2) inds
>         modified = (rands, ts)
>     in  if isOk then modified else (rands, [t1,t2])


This function will drop (make a rest) the first note in a pair. 

> removeFirst :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> removeFirst rands t1 t2 ctxt rulesx = (rands, [(restVal,getDur t1,getMode t1, getKey t1), t2])


This function will drop (make a rest) the second note in a pair

> removeSecond :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> removeSecond rands t1 t2 ctxt rulesx = (rands, [t1, (restVal, getDur t2, getMode t2, getKey t2)])



This function will suspend the first note in a pair accross the second.

> susFirst :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> susFirst rands t1 t2 ctxt rulesx =  
>     let rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail = intersect iAvail1 iAvail2
>         inds = iAvail 
>         isOk = contained (getIndex t1) inds
>         modified = (rands, [changeDur t1 (getDur t1 + getDur t2)])
>     in  if isOk then modified else (rands, [t1,t2])

This function leaves the passage unchanged (just a placeholder for a function list).

> doNothing :: [Int] -> NoteTerm -> NoteTerm -> VContext -> NoteRuleList -> ([Int], [NoteTerm])
> doNothing rands t1 t2 ctxt rules = (rands, [t1,t2])



This function will force an ancitipation and then suspension of the SECOND note.

> susSecond :: [Int] -> NoteTerm -> NoteTerm -> VContext ->  NoteRuleList -> ([Int], [NoteTerm])
> susSecond rands t1 t2 ctxt rulesx = 
>     let rules1 = search (==) rulesx s1
>         rules2 = search (==) rulesx s2
>         s1 = getMode t1
>         s2 = getMode t2
>         k1 = getKey t1
>         k2 = getKey t2
>         ctxt1 = scaleContext ctxt k1
>         ctxt2 = scaleContext ctxt k2
>         iAvail1 = makeIndexSets (getAvailableIndices (ctxt1) rules1)
>         iAvail2 = makeIndexSets (getAvailableIndices (ctxt2) rules2)
>         iAvail = intersect iAvail1 iAvail2
>         inds = iAvail 
>         isOk = contained (getIndex t2) inds
>         modified = (rands, [changeDur t2 (getDur t1 + getDur t2)])
>     in  if isOk then modified else (rands, [t1,t2])




Local type definition for this file only.

> type FModSig = [Int] -> NoteTerm -> NoteTerm -> VContext ->  NoteRuleList -> ([Int], [NoteTerm])


These are unchanging lsits of foreground-adding functions and their probabilities. Ultimately,
these will need to be user-supplied rather than being predefined in this way. 

> fModFuncs = [addPTone, addNToneUp, addNToneDwn, addAnticip, susFirst, susSecond, doNothing]
> fModProbs = [0.25, 	 0.45, 	 0.65, 	  0.85,	  0.90,	0.95,	     1.00] -- modified from foreground.lhs	     
> fModRules = newNoteRules -- repeat myNoteRules



This function chooses a foreground function from a list.

> -- chooseFunB :: [Int] -> [FModSig] -> [[NoteRule]] -> [Rational] -> ([Int], FModSig, NoteRuleList)

> chooseFunB :: [Int] -> [FModSig] -> NoteRuleList -> [Rational] -> ([Int], FModSig, NoteRuleList)
> chooseFunB (r:rs) (f:fs) n (p:ps) = 
>     let r' = (fromIntegral (r `mod` 100)) / 100
>     in  if r' <= p then (rs, f, n) else chooseFunB (r:rs) fs n ps

> chooseFun :: [Int] -> ([Int], FModSig, NoteRuleList)
> chooseFun rands = chooseFunB rands fModFuncs fModRules fModProbs


This function will apply the foreground addition functions, considering notes
as pairs. For example, if a voice has the following structure:

[n1, n2, n3, ...]

The first terms to be considered will be n1 and n2. If these are replaced by

n1', n1'', and n2'

then it will be the case that n1'' is inserted and the next terms that need 
to be considered are n2' and n3. 

> getAllConIndices :: Dur -> Dur -> [Voice2] -> [Int]
> getAllConIndices mind maxd [] = []
> getAllConIndices mind maxd (v:vs) = (getConIndices mind maxd 0 v []) ++ getAllConIndices mind maxd vs




This is the function that adds a complete foreground (one pass) to a voice.

> applyFuns :: [Int] -> Dur -> [NoteTerm] -> [Voice2] -> ([Int], Dur, [NoteTerm])
> applyFuns rands tDur [] voices = (rands, tDur, [])
> applyFuns rands tDur [t] voices = (rands, tDur, [t])
> applyFuns (rands) tDur (t1:t2:ts) voices = 
>     let (rands1, someFun, someRules) = chooseFun rands
>         minD = tDur
>         maxD = tDur + (getDur t1) + (getDur t2)
>         myCon = getAllConIndices minD maxD voices
>         (rands2, newTerms) = someFun rands1 t1 t2 myCon someRules
>         lastTerm = head (reverse newTerms)
>         frontTerms = reverse (tail (reverse newTerms))
>         (rands3, tDur2, ts') = applyFuns rands2 (tDur + getDur t1) (lastTerm : ts) voices
>     in  (rands3, tDur2, frontTerms ++ ts')






============ ADD FOREGROUND - USER INTERFACE ==============

This function assumes that terms are in NOTE form, e.g. that the 
integer lists of indices have only one element. These are 
functions for the user to interact with.

> addForeground :: [NoteTerm] -> [Int] -> ([Int], [NoteTerm])
> addForeground [] rands = (rands, [])
> addForeground [x] rands = (rands, [x])  -- ADDED 25-JAN-2014
> addForeground ts rands = 
>    let (rs', tDur, ts') = applyFuns rands 0 ts [] 
>    in  (rs', ts')

> addConForeground :: [NoteTerm] -> [Int] -> [Voice2] -> ([Int], [NoteTerm])
> addConForeground [] rands voices = (rands, [])
> addConForeground [x] rands voices = (rands, [x]) -- ADDED 25-JAN-2014
> addConForeground ts rands voices =
>    let (rs', tDur, ts') = applyFuns rands 0 ts voices 
>    in  (rs', ts')



> addForegroundsToAll :: [Voice2] -> [Int] -> ([Int], [Voice2])
> addForegroundsToAll [] rands = (rands, [])
> addForegroundsToAll (v:vs) rands = 
>     let (rands2, vs') = addForegroundsToAll vs rands
>         (rands3, v') = addForeground v rands2
>     in  (rands3, v':vs')


These functions handle context-sensitive foregound addition.

> addConForegroundsToAllB :: [Voice2] -> [Int] -> [Voice2] -> ([Int], [Voice2])
> addConForegroundsToAllB [] rands vs = (rands, [])
> addConForegroundsToAllB (v:vs1) rands vs2 = 
>     let (rands1, v') = addConForeground v rands (vs1 ++ vs2)
>         (rands2, vs1') = addConForegroundsToAllB vs1 rands1 (addToEnd v' vs2) 
>     in  (rands2, v':vs1')

> addConForegroundsToAll :: [Voice2] -> [Int] -> ([Int], [Voice2])
> addConForegroundsToAll vs rands = addConForegroundsToAllB vs rands []



============= UTILITIES ================



Data conversion to a Music1 datatype:

> voicesToMusic :: [Voice2] -> Music1
> voicesToMusic [] = rest 0
> voicesToMusic [a] = vtermsToMusic2 a
> voicesToMusic (a:as) = vtermsToMusic2 a :=: voicesToMusic as


> transposeVoice :: AbsPitch -> Voice2 -> Voice2
> transposeVoice a vs = map (\(i,d,s,k) -> (i+a,d,s,k+a)) vs

> voicesToMusicT :: Octave -> AbsPitch -> [Voice2] -> Music1
> voicesToMusicT oct key vs = 
>     let vs' = map (transposeVoice (oct*12+key)) vs
>     in  voicesToMusic vs'

These functions convert Voices to music, as in ToNotes.

> vtermsToMusic2 :: [NoteTerm] -> Music1
> vtermsToMusic2 [] = rest 0
> vtermsToMusic2 [(i,d,_,_)] = 
>     let p = pitch i
>     in  if i<restMax then rest d else note d (p,[])
> vtermsToMusic2 ((i,d,_,_):is) = 
>     let p = pitch i
>         m = if i<restMax then rest d else note d (p,[])
>     in  m :+: vtermsToMusic2 is


> condenseVTerms2 :: [NoteTerm] -> [NoteTerm]
> condenseVTerms2 [] = []
> condenseVTerms2 [a] = [a]
> condenseVTerms2 (a:b:t) = 
>     let dsum = getDur a + getDur b
>         i1 = getIndex a
>         i2 = getIndex b
>         a' = changeDur a dsum
>     in  if i1 == i2 then condenseVTerms2 (a':t) else a : condenseVTerms2 (b:t)


> cReps :: [Int] -> Double -> [NoteTerm] -> [NoteTerm]
> cReps rs thresh [] = []
> cReps rs thresh [x] = [x]
> cReps (r:rs) thresh ((i1, d1, s1, k1):(i2, d2, s2, k2):rest) =
>     if i1==i2 then
>         if ((fromIntegral (r `mod` 100))/100.0 <= thresh) then cReps rs thresh ((i1,d1+d2,s1,k1):rest)
>         else (i1, d1, s1, k1) : cReps rs thresh ((i2, d2, s2, k2):rest)
>     else (i1, d1, s1, k1) : cReps rs thresh ((i2, d2, s2, k2):rest)


> combineReps2 :: Double -> ([Int], Voice2) -> ([Int], Voice2)
> combineReps2 thresh (rands, vs) = (throwAway (length vs) rands, cReps rands thresh vs)


> throwAway :: Int -> [a] -> [a]
> throwAway 0 xs = xs
> throwAway count (h:t) = throwAway (count-1) t


==============

Code taken from NoteRules.lhs in 690.

 module NoteRules where
 import Euterpea
 import GrammarTypes
 import Rules

Donya Quick
Done for: 691 project

This file was mainly a placeholder for further development. Some
basic rules are defined below, but more complex rules are possible.


A note rule consists of the following:
- a list of indices describing a harmonic context
- a list of acceptable additional notes for the context
- a list of probabilities (in cumulative form) for the notes

> defNoteRules :: [NoteRule]
> defNoteRules = [NoteRule([], [], makeEqualProbs 7)]


C D E F G A B
0 1 2 3 4 5 6

> myNoteRules :: [NoteRule]
> myNoteRules = [	NoteRule([0], [0,2,3,4,5], makeEqualProbs 5), 
>			NoteRule([1], [1,3,4,5,6], makeEqualProbs 5), 
>			NoteRule([2], [0,2,4,5,6], makeEqualProbs 5), 
>			NoteRule([3], [0,1,3,4,5], makeEqualProbs 5),  
>			NoteRule([4], [0,1,2,3,4,6], makeEqualProbs 6), 
>			NoteRule([5], [0,1,3,4,5], makeEqualProbs 5),
>			NoteRule([6], [1,2,4,6], makeEqualProbs 4)]



For the next definition, everything will have to be scaled to C for comparison.
This can assume OP-equivalence, but NOT T-equivalence.

	C  C# D  D# E  F  F# G  G# A  A# B  C
Chr:  0  1  2  3  4  5  6  7  8  9  10 11
Maj:  0     1     2  3     4     5     6
Min:  0     1  2     3     4  5     6

> majNoteRules :: [NoteRule]
> majNoteRules = [ NoteRule([0],  [0,4,5,7,9], makeEqualProbs 5), 
>                  NoteRule([1],  [], [0]), 
>                  NoteRule([2],  [2,5,7,9,11], makeEqualProbs 5), 
>                  NoteRule([3],  [], [0]), 
>                  NoteRule([4],  [0,4,7,9,11], makeEqualProbs 4), 
>                  NoteRule([5],  [0,2,5,7,9], makeEqualProbs 5), 
>                  NoteRule([6],  [], [0]), 
>                  NoteRule([7],  [0,2,4,7,11], makeEqualProbs 5), 
>                  NoteRule([8],  [], [0]), 
>                  NoteRule([9],  [0,2,4,5,9], makeEqualProbs 5), 
>                  NoteRule([10], [], [0]), 
>                  NoteRule([11], [2,4,7,11], makeEqualProbs 4) ]

> minNoteRules :: [NoteRule]
> minNoteRules = [ NoteRule([0],  [0,3,5,7,8], makeEqualProbs 5), 
>                  NoteRule([1],  [], [0]), 
>                  NoteRule([2],  [2,5,7,10], makeEqualProbs 5), 
>                  NoteRule([3],  [0,3,5,7,8,10], makeEqualProbs 5),
>                  NoteRule([4],  [], [0]), 
>                  NoteRule([5],  [0,2,3,5,8], makeEqualProbs 4), 
>                  NoteRule([6],  [], [0]), 
>                  NoteRule([7],  [0,2,3,7,10], makeEqualProbs 5),  
>                  NoteRule([8],  [0,3,5,8], makeEqualProbs 5), 
>                  NoteRule([9],  [], [0]),
>                  NoteRule([10], [2,3,7,10], makeEqualProbs 4),
>                  NoteRule([11], [], [0]) ]


> majNR = majNoteRulesStrict
> minNR = minNoteRulesStrict
> newNoteRules = [(Major, majNR), (Minor, minNR)]








> majNoteRulesStrict :: [NoteRule]
> majNoteRulesStrict = [ NoteRule([0],  [0,4,5,7,9], makeEqualProbs 5), 
>                  NoteRule([1],  [], [0]), 
>                  NoteRule([2],  [2,5,7,9,11], makeEqualProbs 5), 
>                  NoteRule([3],  [], [0]), 
>                  NoteRule([4],  [0,4,9,11], makeEqualProbs 4), 
>                  NoteRule([5],  [0,2,5,7,9], makeEqualProbs 5), 
>                  NoteRule([6],  [], [0]), 
>                  NoteRule([7],  [0,2,4,7,11], makeEqualProbs 5), 
>                  NoteRule([8],  [], [0]), 
>                  NoteRule([9],  [0,2,4,5,9], makeEqualProbs 5), 
>                  NoteRule([10], [], [0]), 
>                  NoteRule([11], [2,4,7,11], makeEqualProbs 4) ]

> minNoteRulesStrict :: [NoteRule]
> minNoteRulesStrict = [ NoteRule([0],  [0,3,5,7,8], makeEqualProbs 5), 
>                  NoteRule([1],  [], [0]), 
>                  NoteRule([2],  [2,5,7,10], makeEqualProbs 5), 
>                  NoteRule([3],  [0,3,7,8,10], makeEqualProbs 5),
>                  NoteRule([4],  [], [0]), 
>                  NoteRule([5],  [0,2,5,8], makeEqualProbs 4), 
>                  NoteRule([6],  [], [0]), 
>                  NoteRule([7],  [0,2,3,7,10], makeEqualProbs 5),  
>                  NoteRule([8],  [0,3,5,8], makeEqualProbs 5), 
>                  NoteRule([9],  [], [0]),
>                  NoteRule([10], [2,3,7,10], makeEqualProbs 4),
>                  NoteRule([11], [], [0]) ]



=============

CODE BELOW FROM RULES.LHS IN 690.

Make a set of n equal numbers.

> makeEqualProbsB :: Rational -> Rational -> Int -> [Rational]
> makeEqualProbsB r incr 0 = []
> makeEqualProbsB r incr count = r+incr : makeEqualProbsB (r+incr) incr (count-1)

> makeEqualProbs :: Int -> [Rational]
> makeEqualProbs 0 = []
> makeEqualProbs count = 
>     let incr = 1 / (fromIntegral count)
>     in  makeEqualProbsB 0 incr count


Make a set of numbers cumulative. [1,1,1] becomes [1,2,3].

> makeCumulative :: [Rational] -> [Rational]
> makeCumulative [] = []
> makeCumulative [a] = [a]
> makeCumulative (h:t) = h : makeCumulative (map (+h) t)

> makeRandomProbs :: Int -> Int -> [Rational]
> makeRandomProbs count seed = 
>     let rands = makeInfInts seed
>         vals = map fromIntegral (take count rands)
>         s = sum vals
>         vals' = map (*(1/s)) vals
>         vals'' = makeCumulative vals' 
>     in  vals''


Make infinite random numbers given a seed.

> makeInfInts :: Int -> [Int]
> makeInfInts seed = f (mkStdGen seed) where
>     f g = let (r, g') = next g 
>           in r : f g'