> module GramGen where
> import GrammarMonad hiding (getCtxt)
> import GrammarRules
> import GrammarRulesV
> import System.Random
> import Euterpea hiding (key, dur)
> import ChordSpaces
> import PostProc
> import Data.List
> --import Foreground


> numMajI = 8
> numMajV = 2
> numMinI = 8
> numMinV = 2

> conditions = [((Major, True), numMajI), 
>               ((Major, False), numMajV),
>               ((Minor, True), numMinI),
>               ((Minor, False), numMinV)]
> checkSum = sum $ map snd conditions

> useLets = True
> theRules endOnI = if endOnI then ruleSet minDur useLets
>                   else ruleSetV minDur useLets
> phraseLen = 4 -- four measures
> minDur = qn -- nothing smaller than a quarternote
> maxDur = hn
> minIters = 6 :: Int
> maxIters = 8 :: Int
> tSeed0 = i phraseLen
> tSeed1 = S [i (phraseLen/2), i (phraseLen/2)]
> tSeed2 = Let "x" (i (phraseLen/2)) $ S [Var "x", Var "x"]
> tSeeds = [tSeed0, tSeed1, tSeed2]
> ntLim = 2 -- limit for neighboring tone distance
> ptLim = 3
> pSync = 0.5
> pTie = 0.5
> rootBassThresh = 0.8
> noCPLThresh = 7 -- setting to 0 forces nearest neighbor fallback

Ranges to use? C4 is middle C
Pitch		Euterpea	MIDI (correcting for rooting at C0 vs C1)
S: C4-A5	48-69		60-81
A: E3-E5	40-64		52-76
T: B2-G4	35-55		47-67
B: E2-C4	28-48		40-60


> ranges = [(40,60), (47,67), (52,76), (60,81)]
> allTriads = concatMap f [[0,4,7], [0,3,7], [0,3,6]] where
>     f c = map (\i -> map (+i) c) [0..11]
> allQuadrads = concatMap f [[0,0,4,7], [0,0,3,7], [0,0,3,6], [0,4,7,7], [0,3,7,7]] where
>     f c = map (\i -> map (+i) c) [0..11]
> isChord x = elem (normOPC x) $ map normOPC allQuadrads -- allTriads
> cFilter x = isChord x && x == sort (nub x) && goodSpacing x
> allChords = filter cFilter $ makeRange $ ranges 
> instrs = [Bassoon, EnglishHorn, Clarinet, Oboe]


Other rules to enforce:
http://www.choraleguide.com/vl-spacing.php

> goodSpacing [b,t,a,s] = 
>     s - t <= 19 &&
>     s - a <= 12 && 
>     a - t <= 12 

> qSpace = allChords // opcEq 

> data GPhrase = GPhrase 
>                {pSeed :: Int,
>                 pEnd :: Bool,
>                 pType :: Int,
>                 pKey :: AbsPitch,
>                 pMode :: Mode,
>                 pIters :: Int,
>                 pRep :: Bool,
>                 pTerm :: Term,
>                 pChords :: Music Pitch,
>                 pMusic :: Music Pitch}
>   deriving (Eq, Show)

> makePhrase :: Seed -> Mode -> Bool -> GPhrase
> makePhrase seed m endOnI = 
>     let gs = splitN $ mkStdGen seed -- get all the generators
>         ctxt = if m==Major then 0 else 12
>         i = fst $ randomR (minIters, maxIters) (gs!!0) -- how many iters?
>         t = fst $ randomR (0,2) (gs!!1) -- Term seed? 0=normal, 1=I-I, 2=Let
>         doRep = t>1 -- formalize this as a Bool (True = we repeat)
>         i' = if t>0 then i `div` 2 else i -- correct for generative length
>         k = fst $ randomR (0,11 :: AbsPitch) (gs!!2) -- transpose!
>         tPhrase = doGen endOnI i' (gs!!3) ctxt (tSeeds!!t)-- get Roman numerals
>         csPhrase = toCS k (gs!!4) m tPhrase -- make them into actual chords!
>         vsPhrase = convertTChords csPhrase -- turn note matrix sideways and conver types
>         vMusic = vsToMusicI instrs $ toVoices csPhrase -- what it sounds like as just chords
>         fPhrase = reverse $ snd $ addFG (gs!!5) $ reverse vsPhrase -- add a foreground!
>         fMusic = vsToMusicI instrs $ backToVoices fPhrase -- what it sounds like as just chords
>     in  GPhrase seed endOnI t k m i' doRep tPhrase vMusic fMusic -- return result

Generation continues until the duration distribution is "reasonable." 
In this case, because the duration of the phrase is small, we really want
only halfnotes and quarternotes. Wholenotes sound weird. We want fewer 
generative steps, but if there are still big durations hanging around, they
need to be split up by further generative steps.

> doGen :: Bool -> Int -> StdGen -> Int -> Term -> Term
> doGen end i g ctxt t = recCheck i (gen'' (theRules end) g ctxt t) where
>     recCheck i ts = let tc = ts !! i in
>         if goodDurs tc then tc else recCheck (i+1) ts
>     goodDurs :: Predicate Term
>     goodDurs (Let x a e) = goodDurs a && goodDurs e
>     goodDurs (Var x) = True -- can't tell duration here, so assume ok
>     goodDurs (S xs) = and $ map goodDurs xs
>     goodDurs (Mod m x) = goodDurs x
>     goodDurs (NT (Chord d c)) = d <= maxDur


The "main" function for this module. This does all the work. Given
a number of phrases to write, i, it will write 4i phrases in total 
having the following properties:
- i phrases in Major ending on I from seeds 0 toi-1
- i phrases in Minor ending on I from seeds i to 2i-1
- i phrases in Major ending on V from seeds 2i to 3i-1
- i phrases in Minor ending on V from seeds 3i to 4i-1

> writePhrases :: Int -> IO ()
> writePhrases i = 
>     let seeds = [0..(4*i)-1] :: [Seed]
>         seed = 4*i -- for sample selection
>         modes = take (2*i) (repeat Major) ++ take (2*i) (repeat Minor) :: [Mode]
>         ends = let x = take i (repeat True) ++ take i (repeat False) in x ++ x
>         vals = zipWith3 makePhrase seeds modes ends :: [GPhrase]
>         allMusic = line $ intersperse (rest 2) $ map pMusic vals
>         sample = getSample (mkStdGen seed) vals :: [GPhrase]
>         sampleMusic = line $ intersperse (rest 2) $ map pMusic sample
>     in  sequence (map procGP vals) >> 
>         putStrLn "Writing conglomerate..." >>
>         writeSummaryTxt vals "grammar\\summary.csv" >> 
>         writeMidi "grammar\\grammar_all.mid" allMusic >> 
>         putStrLn "Writing sample...">>
>         writeSummaryTxt sample ("grammar\\sample"++show seed++".csv") >>
>         writeMidi ("grammar\\grammar_sample"++show seed++".mid") sampleMusic >> 
>         putStrLn "Done.\n"

Here a single phrase is processed to write appropriate output files.
Three files are written: a text file with the Term and other info 
about the generated case (key, etc.), a MIDI file of the chordal 
backbone (after chord spaces), and a MIDI file of the final result.

> procGP :: GPhrase -> IO ()
> procGP gp = 
>     let endS = if pEnd gp then "_I" else "_V"
>         fn ext = "grammar\\grammar_"++show (pMode gp)++show (pSeed gp)++endS++ext
>         info = "End on I? "++show (pEnd gp) ++
>                "\nType: "++show (pType gp) ++
>                "\nKey: "++show (pKey gp) ++ " "++ show (pMode gp)++
>                "\nIterations: "++show (pIters gp) ++
>                "\nRepeat: "++show (pRep gp) ++
>                "\n\n"++ showTerm (pTerm gp) ++ "\n\n"
>     in  putStrLn ("Writing example: "++show (pSeed gp)++" "++show (pMode gp)++"...") >>
>         writeFile (fn "_term.txt") info >>
>         writeMidi (fn "_chords.mid") (pChords gp) >>
>         writeMidi (fn "_fg.mid") (pMusic gp)

> writeSummaryTxt :: [GPhrase] -> FilePath -> IO ()
> writeSummaryTxt xs fp = writeFile fp (headers ++ concatMap summarize xs) where
>     headers = "Seed\tEnd Type\tKey\tIters\tRepeats?\n"
>     showEnd b = if b then "I" else "V"
>     showRep b = if b then "Yes" else "No"
>     showType t = ["I", "I-I", "Let"] !! t
>     summarize gp = show (pSeed gp) ++ "\t" ++
>                    showEnd (pEnd gp) ++ "\t" ++
>                    showType (pType gp) ++ "\t" ++
>                    show (pKey gp) ++ " "++ show (pMode gp)++"\t"++
>                    show (pIters gp) ++ "\t" ++
>                    showRep (pRep gp) ++ "\n"




> getSample g examples = concatMap (exTake (randomize g examples)) conditions where
>     exTake exs ((m,endOnI), count) = take count $ 
>         filter (\e -> pEnd e == endOnI && pMode e == m) exs

================

Random handling

> splitN :: StdGen -> [StdGen]
> splitN g = let (g1, g2) = split g in g1 : splitN g2


================

CS step to catch high-level lets

> toCS = topLevelCS

> topLevelCS :: AbsPitch -> StdGen -> Mode -> Term -> [TChord]
> topLevelCS k g m tVal@(Let x xVal xExp) = case xExp of
>     S [Var v1, Var v2] -> if v1==x && v2==x then 
>                               let ret = lowLevelCS k g m (expand [] xVal)
>                               in  ret ++ ret
>                           else lowLevelCS k g m $ expand [] tVal
>     _ -> lowLevelCS k g m $ expand [] tVal
> topLevelCS k g m tVal = lowLevelCS k g m $ expand [] tVal


> lowLevelCS :: AbsPitch -> StdGen -> Mode -> Term -> [TChord]
> lowLevelCS k g m tVal = 
>     let aChords = map (tChordTrans k) $ toAbsChords tVal m
>         kLast = getRoot $ last aChords
>         (durs, cs) = unzip $ map (\(k,d,a) -> ((k,d), a)) aChords
>         ks = map (\((k,m),d) -> k) durs
>         csNew = greedy' ks g cs
>     in  zipWith (\(k,d) a -> (k,d,a)) durs csNew


> tChordTrans :: AbsPitch -> TChord -> TChord
> tChordTrans kOff ((k,m), d, cs) = (((k+kOff) `mod` 12, m), d, t kOff cs)

> getRoot ((k,m),_,_) = k
> basicGreedy = greedyProg qSpace opcEq (noCPL 7) nearFall 

> bassFilter b cs = minimum cs `mod` 12 == b `mod` 12

> stochBass :: StdGen -> [AbsPitch] -> [EqClass AbsChord] -> [EqClass AbsChord]
> stochBass g (b:bs) (e:es) = 
>     let (r,g') = randomR (0,1.0::Double) g
>         e' = if r <= rootBassThresh then filter (bassFilter b) e else e
>     in  e' : stochBass g' bs es

> greedy' ks g cs = 
>     let ecs = map (eqClass qSpace opcEq . normOPC) cs -- get the eq classes
>         lastClass = let x = filter (bassFilter $ last ks) $ last ecs in 
>                     if null x then last ecs else x
>         ecs' = checkEcs cs (take (length cs - 1) ecs ++ [lastClass])
>     in  reverse $ greedy2 g (noCPL noCPLThresh) nearFall $ reverse ecs'

> checkEcs (x:xs) (e:es) =
>     if null e then error ("Empty class for "++show x)
>     else e : checkEcs xs es
> checkEcs _ [] = []


> greedy2 :: (Eq a) => StdGen -> Predicate (a,a) -> Fallback a -> [[a]] -> [a]
> greedy2 g fPred fFall [] = error "(greedy2) Nothing to choose from!"
> greedy2 g fPred fFall xs = 
>     let (i, g') = randomR (0, length (head xs) - 1) g
>     in  if null (head xs) then error "(greedy2) Empty equivalence class!" 
>         else greedyRec2 g fPred fFall (head xs !! i) (tail xs)

> greedyRec2 g fPred fFall x [] = [x]
> greedyRec2 g fPred fFall x (e:ecs) = 
>     let e' = filter (\y -> fPred (x,y)) e
>         (i, g1) = randomR (0, length e' - 1) g
>         (g2, xFall) = fFall e g x
>     in  if null e' then x : greedyRec2 g2 fPred fFall xFall ecs
>         else x : greedyRec2 g1 fPred fFall (e' !! i) ecs




=====================================================================================================

FOREGROUND STUFF BELOW


==============

FG CONVERSION

> data TNote2 = TNote2 {key :: Key,
>                       dur :: Dur,
>                       beat :: Dur,
>                       isLast :: Bool,
>                       ap :: AbsPitch}
>     deriving (Eq, Show, Ord)


> toTN2 :: Dur -> Bool -> TChord -> [TNote2]
> toTN2 o l (k,d,as) = map (\a -> TNote2 k d o l a) as

> toTN1 :: TNote2 -> TNote
> toTN1 t = (key t, dur t, ap t)

> backToVoices :: [[TNote2]] -> [Voice]
> backToVoices = map (map toTN1)

> toTNote2 = convertTChords

> convertTChords :: [TChord] -> [[TNote2]]
> convertTChords tcs = Data.List.transpose $ fRec tcs 0 where
>     fRec [] dSum = []
>     fRec (x@(k,d,as):xs) dSum = toTN2 dSum (null xs) x : fRec xs dSum


===============

FG GENERATION

> baseScale :: Key -> [AbsPitch]
> baseScale (k,m) = normOP $ t k (getScale m)

Get all pitches shared between the scales of two TNotes

> allPs :: TNote2 -> TNote2 -> [AbsPitch]
> allPs t1 t2 = 
>     let o1 = ap t1 `div` 12
>         o2 = ap t2 `div` 12
>         [oMin, oMax] = sort [o1, o2]
>         offs = map (12*) [oMin-1, oMin, oMax, oMax+1]
>         s1 = baseScale $ key t1
>         s2 = baseScale $ key t2
>         s = [s | s<-s1, elem s s2]
>     in  nub $ concatMap (\o -> t o s) offs

> t1 = TNote2 (0,Major) qn 0 False 60
> t2 = TNote2 (0,Major) qn qn False 67


> type ForeFun = StdGen -> TNote2 -> TNote2 -> (StdGen, Maybe AbsPitch)

> pickPT :: AbsPitch -> ForeFun
> pickPT lim g t1 t2 = 
>     let [pMin, pMax] = sort [ap t1, ap t2]
>         f x = x>pMin && x<pMax && (x - pMin <=lim || pMax - x <=lim)
>         ps' = [x | x<-allPs t1 t2, f x]
>         (iNew, g') = randomR (0, length ps' - 1) g
>     in  if pMin == pMax || null ps' then (g, Nothing)
>         else (g', Just $ ps' !! iNew)

> anticip :: ForeFun
> anticip g t1 t2 = (g, Just $ ap t2)

> rept :: ForeFun
> rept g t1 t2 = (g, Just $ ap t1)

> pickNT :: AbsPitch -> ForeFun
> pickNT lim g t1 t2 = 
>     let [pMin, pMax] = sort [ap t1, ap t2]
>         f x = (x < pMin && pMin - x <= lim) || (x > pMax && x - pMax <= lim)
>         ps' = [x | x<-allPs t1 t2, f x]
>         (iNew, g') = randomR (0, length ps' - 1) g
>     in  if pMin == pMax || null ps' then (g, Nothing)
>         else (g', Just $ ps' !! iNew)

> doNothing :: ForeFun
> doNothing g t1 t2 = (g, Nothing)

> f1 = pickPT ptLim
> f2 = pickNT ntLim
> f3 = anticip
> f4 = rept
> f5 = doNothing
> --          1   2   3   4   5   6   7   8   9   10
> foreFuns1 = [f1, f1, f1, f2, f5, f5, f5, f5, f5, f5] :: [ForeFun]
> foreFuns2 = [f1, f1, f1, f5, f5, f5, f5, f5, f5, f5] :: [ForeFun]
> foreFuns3 = [f1, f5, f5, f5, f5, f5, f5, f5, f5, f5] :: [ForeFun]
> foreFuns4 = [f5] :: [ForeFun]

> replaceIth :: [a] -> Int -> a -> [a]
> replaceIth [] i x = error "(replaceIth) Empty list."
> replaceIth xs i x = if i < 0 || i > length xs - 1 then error "(replaceIth) Bad index." else
>     take i xs ++ [x] ++ drop (i+1) xs

> crossPred :: AbsChord -> Int -> AbsPitch -> AbsChord -> Bool
> crossPred chord1 i newP chord2 = 
>     let chord1' = replaceIth chord1 i newP
>     in  hNotCross2 [chord1, chord1', chord2]


> splitP2 :: StdGen -> Int -> AbsPitch -> TNote2 -> (StdGen, [TNote2])
> splitP2 g i newP t = 
>     let (r, g') = randomR (0,1.0::Double) g
>         dOld = dur t
>         dNew = if r < pSync then dur t / 2 else en
>     in  (g', [t{dur = dOld - dNew}, t{dur = dNew, ap = newP}])

> splitP :: Int -> AbsPitch -> TNote2 -> [TNote2]
> splitP i newP t = 
>     let dNew = dur t / 2
>     in  [t{dur = dNew}, t{dur = dNew, ap = newP}]

> -- Here we assume that the contexts include the voice in question. The i argument is the voice number.
> -- The function only returns modifications of the first chord.
> applyForeFun :: StdGen -> Int -> (TNote2, AbsChord) -> (TNote2, AbsChord) -> 
>                 ForeFun -> (StdGen, [TNote2])
> applyForeFun g i (t1, ctxt1) (t2, ctxt2) fFun = 
>     let (g1, newP) = fFun g t1 t2
>     in  case newP of Nothing -> (g1, [t1])
>                      Just x -> splitP2 g1 i x t1 -- if crossPred ctxt1 i x ctxt2 then (g', splitP i x t1)
>                                -- else (g', [t1]) -- couldn't add without violations

> addFgToVoice :: [ForeFun] -> StdGen -> Int -> [TNote2] -> [[TNote2]] -> (StdGen, [TNote2])
> addFgToVoice foreFuns g i (t1:t2:ts) allVs = -- pair case
>     let ctxt1 = [] -- getCtxts1 (beat t1 + dur t1) allVs 
>         ctxt2 = [] -- getCtxts2 (beat t2) allVs
>         (j, g1) = randomR (0, length foreFuns - 1) g
>         fFun = foreFuns !! j
>         (g2, t1') = applyForeFun g1 i (t1, ctxt1) (t2, ctxt2) fFun
>         (g3, tRest) = addFgToVoice foreFuns g2 i (t2:ts) allVs
>     in  (g3, t1' ++ tRest)
> addFgToVoice foreFuns g i x allVs = (g, x)

> getCtxt :: Dur -> [TNote2] -> [AbsPitch]
> getCtxt dOn [] = []
> getCtxt dOn (t:ts) = 
>     let endsAt = beat t + dur t
>     in  if (beat t <=dOn && endsAt >= dOn) -- starts before or at same, ends after or at same
>         then ap t : getCtxt dOn ts 
>         else getCtxt dOn ts 

> getCtxt1 :: Dur -> [TNote2] -> [AbsPitch]
> getCtxt1 dCheck [] = []
> getCtxt1 dCheck (t:ts) = 
>     let endsAt = beat t + dur t
>     in  if (beat t < dCheck && endsAt == dCheck) -- for first chord
>         then ap t : getCtxt1 dCheck ts 
>         else getCtxt1 dCheck ts 

> getCtxt2 :: Dur -> [TNote2] -> [AbsPitch]
> getCtxt2 dCheck [] = []
> getCtxt2 dCheck (t:ts) = 
>     let endsAt = beat t + dur t
>     in  if (beat t == dCheck && endsAt > dCheck) -- for second chord
>         then ap t : getCtxt2 dCheck ts 
>         else getCtxt2 dCheck ts 

> getCtxts :: Dur -> [[TNote2]] ->  [AbsPitch]
> getCtxts dOn = concatMap (getCtxt dOn)

> getCtxts1 :: Dur -> [[TNote2]] ->  [AbsPitch]
> getCtxts1 dCheck = concatMap (getCtxt1 dCheck)

> getCtxts2 :: Dur -> [[TNote2]] ->  [AbsPitch]
> getCtxts2 dCheck = concatMap (getCtxt2 dCheck)

> allFFs = [foreFuns1, foreFuns2, foreFuns3, foreFuns4]

> addFG :: StdGen -> [[TNote2]] -> (StdGen, [[TNote2]])
> addFG g vs = let (g', vs') = fgRec g 0 vs in tieRec g' vs' where
>     fgRec g i vs = if i >= length vs then (g, vs) else -- CHANGE THIS
>         let (g', v') = addFgToVoice (allFFs !! i) g i (vs !! i) vs 
>             vs' = replaceIth vs i v'
>         in  fgRec g' (i+1) vs'
>     tieRec g [] = (g, [])
>     tieRec g (v:vs) = 
>         let (g1,v') = stochTie g v
>             (g2,vs') = tieRec g1 vs
>         in  (g2, v':vs')


> stochTie :: StdGen -> [TNote2] -> (StdGen, [TNote2])
> stochTie g (t1:t2:ts) = 
>     let (r, g1) = randomR (0, 1.0::Double) g
>         (g2, (t2':ts')) = stochTie g1 (t2:ts)
>         (d1,d2') = (dur t1, dur t2')
>     in  if ap t1 == ap t2' && r < pTie then (g2, t1 {dur = d1 + d2'} : ts')
>         else (g2, t1 : t2' : ts')
> stochTie g ts = (g, ts)

=============================================

UTILITY FUNCTIONS

> infInts g = let (i,g') = next g in i : infInts g'
> toVs2 = map (map f) where f ((k,m),d,p) = (p, d, m, k)
> toVs1 = map (map f) where f (p,d,m,k) = ((k,m),d,p)

