Harmony Grammar Monad Implementations 
Donya Quick and Paul Hudak
Last modified: 19-June-2013

WORK IN PROGRESS

> module GrammarMonad where
> import Euterpea
> import Data.List
> import Control.Monad.State
> import System.Random

Construction of the grammar's symbols and structure

A CType is the Roman numeral.

> data CType = I | II | III | IV | V | VI | VII
>      deriving (Eq, Show, Ord, Enum)

An MType is a degree of modulation (relative 2nd, 3rd, ..., 7th). 
There is no M1 since that would indicate remaining in the home key.

> data MType = M2 | M3 | M4 | M5 | M6 | M7
>      deriving (Eq, Show, Ord, Enum)

A Chord is a duration and CType.

> data Chord = Chord Dur CType
>      deriving (Eq, Show)

The Term data structure has five constructors: a chord, which acts as
a nonterminal but can also be a terminal; a "sentence," or sequence of
Terms; a modulation applied to a Term; a variable; and a Let statement
that uses variables that are represented as strings.

> data Term = NT Chord | S [Term] | Mod MType Term |  
>             Let String Term Term | Var String
>      deriving (Eq, Show)

> data Rule = (CType, Prob) :-> RuleFun

> type RuleFun = Ctxt -> Dur -> Term
> type Prob = Double

These are used to simplify writing rules.

> [i, ii, iii, iv, v, vi, vii] = map (\c t -> NT (Chord t c)) $ enumFrom I 
> c ct t = NT (Chord t ct)
> rf fs t = map ($t) fs

======== CONTEXT ========

The following data type will represent a state, which contains a 
list of Terms and the current random number seed.

> type Ctxt = Int 

> type CtxtFun = Ctxt -> MType -> Ctxt 


> --        d  e  f  g  a  b
> majPatM = [1, 1, 0, 0, 1, 1] -- 1 is minor
> minPatM = [1, 0, 1, 1, 0, 0] -- 1 is minor
> majPatK = [2,4,5,7,9,11]
> minPatK = [2,3,5,7,8,10] -- BUG FIX 25-jAN-2014: missing 7 from list

> defCtxtFun :: CtxtFun
> defCtxtFun cOld m = 
>     let i = fromEnum m
>         (s,o) = if cOld < 12 then (majPatK, majPatM) else (minPatK, minPatM)
>     in  if i >= 6 then error ("(defCtxtFun) Bad modulation: "++show i) 
>         else (s !! i) + 12 * (o !! i)


=========== MONAD IMPLEMENTATION ============

> newtype Prog a = Prog ((StdGen, Ctxt) -> ((StdGen, Ctxt), a))

Define Prog as an instance of Monad:

> instance Monad Prog where
>     return a       = Prog (\s -> (s,a))
>     Prog p0 >>= f1 = Prog $ \s0 ->
>         let (s1, a1) = p0 s0
>             Prog p1 = f1 a1
>         in  p1 s1

One domain-specific primitive:

> getRand :: (Random a) => (a, a) -> Prog a
> getRand ran = Prog (\(g,c) -> let (r,g') = randomR ran g
>               in ((g',c),r))

> getCtxt :: Prog Ctxt
> getCtxt = Prog (\(g,c) -> ((g,c),c))

> setCtxt :: Ctxt -> Prog ()
> setCtxt c' = Prog (\(g,c) -> ((g,c'), ())) 

To "run" the Prog monad:

> runP :: Prog a -> StdGen -> Ctxt -> a
> runP (Prog f) g c = snd (f (g,c))

-------------------------------------------------------------------------

A function to rewrite one term to another:

> update :: [Rule] -> Term -> Prog Term
> update rules t = case t of
>     NT x      -> applyRule rules x
>     S s       -> do ss <- sequence (map (update rules) s)
>                     return (S ss)
>     Mod m s   -> do c <- getCtxt
>                     setCtxt (defCtxtFun c m) -- into mod
>                     s' <- update rules s
>                     setCtxt c -- out of mod
>                     return (Mod m s')
>     Var x     -> return (Var x)
>     Let x a t -> do a' <- update rules a
>                     t' <- update rules t
>                     return (Let x a' t')

> applyRule :: [Rule] -> Chord -> Prog Term
> applyRule rules (Chord d c) =  
>   let  rs = filter (\((c',p) :-> rf) -> c'==c) rules
>   in   do r <- getRand (0.0, 1.0)
>           x <- getCtxt -- NEED TO DO SOMETHING WITH THIS?
>           return ((choose rs r) x d)

> choose :: [Rule] -> Prob -> RuleFun
> choose [] p  = error "Nothing to choose from!"
> choose (((c,p') :-> rf):rs) p  = 
>     if p<=p' || null rs then rf else choose rs (p-p') 

Iterate a monadic function a given number of times:

> iter :: Monad m => (a -> m a) -> a -> m [a]
> iter f a = do a' <- f a
>               as <- iter f a' 
>               return (a':as)

-----------------------------------------------------------------------

Putting it all together:

> gen :: [Rule] -> Int -> Seed -> Ctxt -> Term -> Term
> gen rules i s c t = runP (iter (update rules) t) (mkStdGen s) c !! i

> gen' :: [Rule] -> Int -> StdGen -> Ctxt -> Term -> Term
> gen' rules i s c t = runP (iter (update rules) t) s c !! i

> gen'' :: [Rule] -> StdGen -> Ctxt -> Term -> [Term] -- gives all results
> gen'' rules s c t = runP (iter (update rules) t) s c

> type Seed = Int

The expand function eliminates Lets and Vars from a generated term.
It allows for nested Let expressions for variables with the same name
with lexical scoping. For example:

expand [] (Let "x" t1 (Let "x" t2 (Var "x"))) ==> t2

> expand :: [(String, Term)] -> Term -> Term 
> expand e t = case t of 
>     Let x a exp -> expand ((x, expand e a):e) exp
>     Var x       -> maybe (error (x ++ " is undefined")) id $ lookup x e
>     S s         -> S $ map (expand e) s
>     Mod m t'    -> Mod m $ expand e t'
>     x           -> x
