Rule utilities for use with GrammarRules.lhs (and other rule modules)
Donya Quick and Paul Hudak
Last modified: 12-June-2013
For paper: Grammar-Based Automated Music Composition in Haskell

This module adds some utility functions for use in defining rule
sets. This includes "shorthand" functions for writing single chord
Term values.

> module RuleUtils where
> import Euterpea
> import GrammarMonad


Rule probabilities need to sum to 1.0 for rules with the same lefthand 
side. Since there is an option to include or exclude the let-in rules, 
the probabilities should be normalized (forced to sum to 1.0) before 
the rules are used.

> lhs ((c,p) :-> rf) = c
> prob ((c,p) :-> rf) = p

> normalize :: [Rule] -> [Rule]
> normalize [] = []
> normalize (r@((l,p) :-> rf):rs) = 
>     let rset = r : filter ((l==).lhs) rs
>         rset' = filter ((l/=).lhs) rs
>         psum = sum $ map prob rset
>     in  map (\((l',p') :-> c') -> ((l',p'/psum) :-> c')) rset ++ normalize rset'

Rules can be "wrapped" to produce rules that only operate on chords
with at least a certain duration. 

> toRelDur :: Dur -> Rule -> Rule
> toRelDur d ((c,p) :-> f) = 
>     let dmin ctxt t = minDur $ expand [] $ f ctxt t
>     in  ((c, p) :-> \ctxt t -> if dmin ctxt t < d then NT $ Chord t c else f ctxt t)

> minDur :: Term -> Dur
> minDur (S s) = minimum $ map minDur s
> minDur (Mod m t) = minDur t
> minDur (NT (Chord d x)) = d
> minDur _ = error "(minDur) String is not fully interpreted."

==== PRINTING FUNCTIONS ====

> showRule :: Rule -> String
> showRule ((l,p) :-> rf) = show p ++ "\t" ++ show l ++ " -> " ++ show (rf 0 1.0) ++ "\n"

> writeRules :: [Rule] -> IO ()
> writeRules ruleSet = 
>     let rstr = concatMap showRule ruleSet 
>     in  writeFile "rules.txt" rstr